#-----------------------------------------------------------------------------------
#
#   SCRIPT      attachObjects.py
#   AUTHOR      Evgen Zaitsev
#               ev.zaitsev@gmail.com
#
#   3dsmax style Attach polygons.
#   Attaches objects to the last selected object.
#   Skinning supported, uses polyAppendVertex and other maya.cmds and maya.OpenMaya commands for transfering data
#
#   VERY SLOW
#   
#-----------------------------------------------------------------------------------
'''
The same like in 3dsmax. Attach polygons via maya python api and cmds to existing mesh. 
Not combine, no cheating. Also with skinning support. 

 USAGE:
 
import sys
import maya.cmds as cmds 
path = 'C:/maya_scripts/maya/modeling'
if path not in sys.path:
    sys.path.append(path) 
import attachObjects; attachObjects.main();

'''
import maya.cmds as cmds
import maya.OpenMaya        as OpenMaya     # python API 1.0
import time

def main():
    m_atObj = attachObject()
    del m_atObj

class attachObject( object ):
    def __init__( self ):
        self.m_target       = OpenMaya.MDagPath()
        self.m_sources      = []
        self.m_curr         = OpenMaya.MDagPath()
        self.m_faceItems    = []
        self.runOnSelected()
        
    def runOnSelected( self ):
        self.m_sources = []
        # iterate over selected objects
        m_list = OpenMaya.MSelectionList() 
        OpenMaya.MGlobal.getActiveSelectionList( m_list )
        m_listIt = OpenMaya.MItSelectionList( m_list ) 
        while not m_listIt.isDone():
            m_path = OpenMaya.MDagPath()   # will hold a path to the selected object 
            if (OpenMaya.MItSelectionList.kDagSelectionItem == m_listIt.itemType()):
                m_listIt.getDagPath( m_path )
                if ( m_path.hasFn( OpenMaya.MFn.kMesh )):
                    self.m_sources.append(m_path)
            # if the component is list is valid
            m_listIt.next()
        if (len(self.m_sources)>1):
            self.m_target = self.m_sources.pop()
            # collecting required attributes to transfer
            for m_path in self.m_sources:
                self.m_curr = m_path
                self.runItem()
            cmds.select( clear=True )
            cmds.select( self.m_target.fullPathName(), r=True )
        else:
            OpenMaya.MGlobal.displayWarning("Select: 1st - source, 2nd - target")
            
    def runItem( self ):
        OpenMaya.MGlobal.displayInfo("-----------------------")
        cmds.bakePartialHistory(self.m_target.fullPathName(),prePostDeformers=True)
        OpenMaya.MGlobal.displayInfo("{0:<25} {1} to {2}".format("Attaching",self.m_curr.fullPathName(),self.m_target.fullPathName()))
        m_startTime = time.time()
        #-----------------
        self.m_faceItems = []
        m_fnMesh = OpenMaya.MFnMesh( self.m_curr )
        #-----------------
        m_fnMeshTarget = OpenMaya.MFnMesh( self.m_target )
        m_fnDagNodeTarget = OpenMaya.MFnDagNode( self.m_target )
        m_polygonIndex = m_fnMeshTarget.numPolygons()
        m_matrix = m_fnDagNodeTarget.transformationMatrix() 
        #----------------
        m_itPoly = OpenMaya.MItMeshPolygon( self.m_curr )
        while not m_itPoly.isDone():
            m_polygonIndexSource  = m_itPoly.index()
            m_polygonVertices     = OpenMaya.MIntArray()
            m_polygonEdges        = OpenMaya.MIntArray()
            m_itPoly.getVertices( m_polygonVertices )
            m_itPoly.getEdges( m_polygonEdges )
            # generate faceArray
            m_faceArray = []
            m_posS      = []
            #----------------
            for m_polyVertex in m_polygonVertices:
                m_point = OpenMaya.MPoint()
                m_fnMesh.getPoint( m_polyVertex, m_point, OpenMaya.MSpace.kWorld )
                m_posS.append("{:.2f} {:.2f} {:.2f}".format(m_point.x,m_point.y,m_point.z))
                # compute correct point position
                m_point = computeTransform(m_point,self.m_target )
                #m_point = m_point*m_matrix.inverse()
                m_faceArray.append((m_point.x,m_point.y,m_point.z))
            # add polygon
            cmds.polyAppendVertex("{}".format(self.m_target.fullPathName()),a=m_faceArray,ch=False)
            # -------------------
            #m_polygonVertSorted = self.findPolygonVertices(m_polygonVertices,m_posS,m_polygonIndex)
            m_indsOrig, m_indsSorted, m_locIds = self.sortPolygonVertices(m_polygonIndex)
            #print("m_indsOrig, m_indsSorted, m_locIds",m_indsOrig, m_indsSorted, m_locIds)
            #----------------
            self.m_faceItems.append( faceItem(  m_polygonIndex, 
                                                m_polygonIndexSource, 
                                                m_polygonVertices, 
                                                m_polygonEdges,
                                                m_indsSorted,
                                                m_locIds ) )
            #----------------
            m_polygonIndex += 1
            m_itPoly.next()
        m_fnMeshTarget.updateSurface()
        m_endTime = time.time()
        OpenMaya.MGlobal.displayInfo("{0:<25} {1:16.4f} seconds".format("Append polygons in",m_endTime-m_startTime))
        #----------------
        cmds.bakePartialHistory(self.m_target.fullPathName(),prePostDeformers=True)
        self.assignMaterials()
        #----------------
        cmds.bakePartialHistory(self.m_target.fullPathName(),prePostDeformers=True)
        self.copyUVs()
        #----------------
        cmds.bakePartialHistory(self.m_target.fullPathName(),prePostDeformers=True)
        self.copyVertexColors()
        #----------------
        cmds.bakePartialHistory(self.m_target.fullPathName(),prePostDeformers=True)
        self.mergeVertices()
        #----------------
        cmds.bakePartialHistory(self.m_target.fullPathName(),prePostDeformers=True)
        self.setEdgeSmooth()
        #----------------
        cmds.bakePartialHistory(self.m_target.fullPathName(),prePostDeformers=True)
        self.mergeUVs()
        #----------------
        cmds.bakePartialHistory(self.m_target.fullPathName(),prePostDeformers=True)
        m_fnMeshTarget.updateSurface()
        
    def findPolygonVertices( self, m_indsS, m_posS , m_polygonIndex ):
        m_fnMesh = OpenMaya.MFnMesh( self.m_target )
        m_polygonVertices = OpenMaya.MIntArray()
        m_fnMesh.getPolygonVertices(m_polygonIndex,m_polygonVertices)
        m_indsT = []
        m_posT  = []
        for m_polyVertex in m_polygonVertices:
            m_point = OpenMaya.MPoint()
            m_fnMesh.getPoint( m_polyVertex, m_point, OpenMaya.MSpace.kWorld )
            m_posT.append("{:.2f} {:.2f} {:.2f}".format(m_point.x,m_point.y,m_point.z))
        m_indsT = m_polygonVertices
        #----------------
        m_indSortedT = []
        for i in range(len(m_indsS)):
            m_indexS    = m_indsS[i]
            m_positionS = m_posS[i]
            m_arrayIndex = m_posT.index(m_positionS)
            m_indSortedT.append(m_indsT[m_arrayIndex])
        #print(m_indsT)
        return m_indSortedT
        
    def sortPolygonVertices( self, m_polygonIndex ):
        m_fnMesh = OpenMaya.MFnMesh( self.m_target )
        m_polygonVertices = OpenMaya.MIntArray()
        m_fnMesh.getPolygonVertices(m_polygonIndex,m_polygonVertices)
        m_indsOrig   = []
        m_indsSorted = []
        for m_ind in m_polygonVertices:
            m_indsOrig.append(m_ind)
            m_indsSorted.append(m_ind)
        #----------------
        m_indsSorted.sort()
        m_locIds = []
        m_localVertId = 0
        for m_ind in m_indsSorted:
            m_locIds.append(m_indsOrig.index(m_ind))
            m_localVertId += 1
        #----------------
        return m_indsOrig, m_indsSorted, m_locIds
            
    def assignMaterials( self ):
        # TODO: optimize cmd.sets to run it in 1 go like with polySoftEdge
        m_startTime = time.time()
        m_tagertSGs = getFaceListWithShadingGroupsFromMDagPath( self.m_curr )
        for m_faceItem in self.m_faceItems:
            m_faceItem.s_matSG = m_tagertSGs[m_faceItem.s_polyIndex]
            # assign propper shadingGroup
            cmds.sets( "{}.f[{}]".format(self.m_target.fullPathName(),m_faceItem.m_polyIndex), 
                        edit  = True, 
                        forceElement = m_faceItem.s_matSG )
        m_endTime = time.time()
        OpenMaya.MGlobal.displayInfo("{0:<25} {1:16.4f} seconds".format("Assign shadingGroups in",m_endTime-m_startTime))
        
    def copyUVs( self ):
        m_startTime = time.time()
        #-----------------
        m_fnMeshT = OpenMaya.MFnMesh( self.m_target )
        m_fnMeshS = OpenMaya.MFnMesh( self.m_curr )
        m_listT = []
        m_listS = []
        m_fnMeshT.getUVSetNames( m_listT )
        m_fnMeshS.getUVSetNames( m_listS )
        for i in range(len(m_listT)):
            m_uvSetT = m_listT[i]
            m_continue = True
            if i < len(m_listS):
                m_uvSetS   = m_listS[i]
            else:
                m_uvSetS = m_listS[0]
                m_continue = False
            #print(i,m_uvSetT,m_uvSetS)
            if (m_continue):
                cmds.polyUVSet( self.m_target.fullPathName(), currentUVSet=True, uvSet=m_uvSetT)
                for m_faceItem in self.m_faceItems:
                    m_localVertId = 0
                    for m_locId in m_faceItem.m_locIds:
                        m_idS,m_uS,m_vS = getPolygonUV(m_fnMeshS,m_faceItem.s_polyIndex,m_localVertId,m_uvSetS)
                        m_idT,m_uT,m_vT = getPolygonUV(m_fnMeshT,m_faceItem.m_polyIndex,m_locId,m_uvSetT)
                        #print("{}.map[{}]".format( self.m_target.fullPathName(),m_idT))
                        #print(m_uS,m_vS,m_uvSetT)
                        
                        cmds.polyEditUV("{}.map[{}]".format( self.m_target.fullPathName(),m_idT),
                                            relative=False,
                                            uValue=float(m_uS), 
                                            vValue=float(m_vS),
                                            uvs=m_uvSetT)
                        m_localVertId +=1
        #-----------------
        m_endTime = time.time()
        OpenMaya.MGlobal.displayInfo("{0:<25} {1:16.4f} seconds".format("Copy UV's in",m_endTime-m_startTime))
        
    def copyVertexColors( self ):
        m_startTime = time.time()
        #-----------------
        m_fnMeshT = OpenMaya.MFnMesh( self.m_target )
        m_fnMeshS = OpenMaya.MFnMesh( self.m_curr )
        m_listT = []
        m_listS = []
        m_fnMeshT.getColorSetNames( m_listT )
        m_fnMeshS.getColorSetNames( m_listS )
        for i in range(len(m_listT)):
            m_colorSetT = m_listT[i]
            m_continue = True
            if i < len(m_listS):
                m_colorSetS   = m_listS[i]
            else:
                m_colorSetS = m_listS[0]
                m_continue = False
            if (m_continue):
                #-----------------
                m_vertexColours  = OpenMaya.MColorArray()
                m_fnMeshS.getFaceVertexColors(m_vertexColours,m_colorSetS)
                #-----------------
                u_util = OpenMaya.MScriptUtil()
                u_util.createFromInt(0)
                m_ptr = u_util.asIntPtr()
                #-----------------
                cmds.polyColorSet( self.m_target.fullPathName(), currentColorSet=True, colorSet=m_colorSetT)
                for m_faceItem in self.m_faceItems:
                    m_localVertId = 0
                    for m_polyVert in m_faceItem.m_polyVertices:
                        m_fnMeshS.getFaceVertexColorIndex(m_faceItem.s_polyIndex,m_localVertId,m_ptr,m_colorSetS)
                        m_colorId = u_util.getIntArrayItem(m_ptr,0)
                        m_color = m_vertexColours[m_colorId]
                        # vtx  - global vertex id 
                        # Face - global face id
                        cmds.polyColorPerVertex(    "{}.vtxFace[{}][{}]".format( self.m_target.fullPathName(),
                                                                                 m_polyVert,
                                                                                 m_faceItem.m_polyIndex ), 
                                                    rgb=(m_color.r, m_color.g, m_color.b) )
                        m_localVertId +=1
        if len(m_listT)>0:
            cmds.polyColorSet( self.m_target.fullPathName(), currentColorSet=True, colorSet=m_listT[0])
        #-----------------
        m_endTime = time.time()
        OpenMaya.MGlobal.displayInfo("{0:<25} {1:16.4f} seconds".format("Copy VertexColors's in",m_endTime-m_startTime))
        
    def mergeUVs( self ):
        m_startTime = time.time()
        #-----------------
        m_fnMeshT = OpenMaya.MFnMesh( self.m_target )
        m_listT = []
        m_fnMeshT.getUVSetNames( m_listT )
        m_listT.reverse()
        for m_uvSet in m_listT:
            cmds.polyUVSet( self.m_target.fullPathName(), currentUVSet=True,  uvSet=m_uvSet)
            cmds.polyMergeUV("{}.e[*]".format( self.m_target.fullPathName()), distance=0.0001, uvSetName=m_uvSet)
        #-----------------
        m_endTime = time.time()
        OpenMaya.MGlobal.displayInfo("{0:<25} {1:16.4f} seconds".format("Merge UV's in",m_endTime-m_startTime))
        
    def mergeVertices( self ):
        m_startTime = time.time()
        m_vertices = []
        for m_faceItem in self.m_faceItems:
            m_vertices += m_faceItem.m_polyVertices
        m_vertices = list(set(m_vertices))
        m_vertices.sort()
        # merge vertices
        cmds.polyMergeVertex("{}.vtx[{}:{}]".format(self.m_target.fullPathName(),m_vertices[0],m_vertices[-1]),
                                d=0.0001,ch=False,am=True )
        m_endTime = time.time()
        OpenMaya.MGlobal.displayInfo("{0:<25} {1:16.4f} seconds".format("Merge vertices in",m_endTime-m_startTime))
                                
    def setEdgeSmooth( self ):
        m_startTime = time.time()
        # poistion in m_faceItem array == to source face id
        m_itPoly = OpenMaya.MItMeshPolygon( self.m_target )
        #-----------------
        m_util = OpenMaya.MScriptUtil()
        m_util.createFromInt(0)
        m_prevIndex  = m_util.asIntPtr()
        #-----------------
        m_edges = []
        for m_faceItem in self.m_faceItems:
            m_polygonEdges = OpenMaya.MIntArray()
            m_itPoly.setIndex( m_faceItem.m_polyIndex, m_prevIndex )
            m_itPoly.getEdges( m_polygonEdges )
            m_faceItem.m_polygonEdges = m_polygonEdges
            #print(m_faceItem.m_polyIndex,m_faceItem.s_polyIndex)
            #print(m_faceItem.s_polygonEdges,m_faceItem.m_polygonEdges)
            m_edges += m_faceItem.m_polygonEdges
        m_edges = list(set(m_edges))
        m_edges.sort()
        # Collect data about target edges
        #print(m_edges)
        m_edgesDict = {}
        m_facesEdge = {}
        m_itEdgeIt  = OpenMaya.MItMeshEdge( self.m_target ) 
        m_fnMesh    = OpenMaya.MFnMesh( self.m_target )
        for m_edgeId in m_edges:
            m_facesArray = OpenMaya.MIntArray()
            m_itEdgeIt.setIndex( m_edgeId, m_prevIndex )
            m_itEdgeIt.getConnectedFaces(m_facesArray) 
            if 2==len(m_facesArray):
                m_facesEdge[ (m_facesArray[0],m_facesArray[1]) ] = m_edgeId
                m_facesEdge[ (m_facesArray[1],m_facesArray[0]) ] = m_edgeId
            #print(m_edgeId,m_facesArray)
            m_edgesDict[m_edgeId] = False
        #print(m_edgesDict)
        #print(m_facesEdge)
        #-----------------
        # Go through all source edges and get smoothing
        m_itEdgeIt = OpenMaya.MItMeshEdge( self.m_curr ) 
        m_fnMesh   = OpenMaya.MFnMesh( self.m_curr )
        while not m_itEdgeIt.isDone():
            m_facesArray = OpenMaya.MIntArray()
            m_edgeId = m_itEdgeIt.index()
            m_itEdgeIt.getConnectedFaces(m_facesArray) 
            #-----------------
            m_start, m_end = getEdgeVertices(m_fnMesh,m_edgeId) 
            m_edgeSmooth   = isEdgeSmooth(m_fnMesh,m_start,m_end,m_facesArray)
            #print( m_edgeId,m_facesArray,(m_start, m_end), m_edgeSmooth )
            m_edgeIdTarget = None
            # process edges only with 2 polygons connected,
            # all others will be set to hardEdge
            if 2==len(m_facesArray):
                m_f1 = (self.m_faceItems[m_facesArray[0]].m_polyIndex,self.m_faceItems[m_facesArray[1]].m_polyIndex)
                m_f2 = (self.m_faceItems[m_facesArray[1]].m_polyIndex,self.m_faceItems[m_facesArray[0]].m_polyIndex)
                if   m_facesEdge.has_key(m_f1):
                    m_edgeIdTarget = m_facesEdge[m_f1]
                elif m_facesEdge.has_key(m_f2):
                    m_edgeIdTarget = m_facesEdge[m_f2]
            if (m_edgeIdTarget):
                m_edgesDict[m_edgeIdTarget] = m_edgeSmooth
            #print( m_edgeId,m_facesArray,m_edgeIdTarget,m_edgeSmooth )
            m_itEdgeIt.next()
        #-----------------
        # isolate only soft edges
        m_softEdges = []
        for m_key in m_edgesDict.keys():
            if m_edgesDict[m_key]:
                m_softEdges.append(m_key)
        m_softEdges.sort()
        #print(m_softEdges)
        #-----------------
        if len(m_softEdges)>0:
            self.setSoft(m_softEdges)
        m_endTime = time.time()
        OpenMaya.MGlobal.displayInfo("{0:<25} {1:16.4f} seconds".format("Set soft/hard edges in",m_endTime-m_startTime))
        
    def setSoft( self, m_softEdges ):
        # select and set Hard Edges 
        m_aMember           = ''
        m_lastIndices       = [ -1, -1 ]
        m_haveEdge          = False
        for m_edgeId in m_softEdges:
            if ( -1 == m_lastIndices[0] ):
                m_lastIndices[0] = m_edgeId
                m_lastIndices[1] = m_edgeId
            else:
                m_currentIndex = m_edgeId
                if ( m_currentIndex > m_lastIndices[1] + 1 ):
                    m_aMember += '{0}.e[{1}:{2}] '.format( self.m_target.fullPathName(), m_lastIndices[0], m_lastIndices[1] )
                    m_lastIndices[0] = m_currentIndex
                    m_lastIndices[1] = m_currentIndex 
                else:
                    m_lastIndices[1] = m_currentIndex
            m_haveEdge = True
        if ( m_haveEdge ):
            m_aMember += '{0}.e[{1}:{2}] '.format( self.m_target.fullPathName(), m_lastIndices[0], m_lastIndices[1] )
        m_str = ""
        m_str += "select -r {0};\n".format( m_aMember )
        m_str += "polySoftEdge -a 180 -ch 1;\n"
        m_str += "select -cl;"
        #print(m_str)
        OpenMaya.MGlobal.executeCommand( m_str )
        
    def runItemAddPolygon( self ):
        # not stable implementation
        # do not work with skinning and has no hsitory 
        m_fnMesh = OpenMaya.MFnMesh( self.m_curr )
        #-----------------
        m_fnMeshTarget = OpenMaya.MFnMesh( self.m_target )
        m_fnDagNodeTarget = OpenMaya.MFnDagNode( self.m_target )
        m_matrix = m_fnDagNodeTarget.transformationMatrix() 
        #----------------
        m_itPoly = OpenMaya.MItMeshPolygon( self.m_curr )
        while not m_itPoly.isDone():
            m_polygonIndexSource  = m_itPoly.index()
            m_polygonVertices     = OpenMaya.MIntArray()
            m_itPoly.getVertices( m_polygonVertices )
            print(m_polygonIndexSource,m_polygonVertices)
            # generate faceArray
            m_faceArray = OpenMaya.MPointArray()
            m_faceArray.setLength(len(m_polygonVertices))
            m_l = 0
            # - pointer
            m_util = OpenMaya.MScriptUtil()
            m_util.createFromInt(0)
            m_faceIndexPtr = m_util.asIntPtr()
            # ----
            for m_polyVertex in m_polygonVertices:
                m_point = OpenMaya.MPoint()
                m_fnMesh.getPoint( m_polyVertex, m_point, OpenMaya.MSpace.kWorld )
                m_faceArray.set(m_point*m_matrix.inverse(),m_l)
                m_l += 1
            # add polygon
            m_result = m_fnMeshTarget.addPolygon( m_faceArray, m_faceIndexPtr, True, 0.0001, self.m_curr.node() )
            print(m_util.getInt(m_faceIndexPtr))
            # --------
            m_itPoly.next()
            
class faceItem( object ):
    def __init__( self, m_polyIndex, s_polyIndex, s_polyVertices, s_polygonEdges, m_polyVertices, m_locIds ):
        self.m_polyIndex    = m_polyIndex               # target polygon index
        self.m_polyVertices = m_polyVertices            # target polygon vertices
        self.m_locIds       = m_locIds                  # list of localIds ot the face
        self.m_polygonEdges = []                        # target polygon edges
        self.s_polyIndex    = s_polyIndex               # source polygon index
        self.s_polyVertices = s_polyVertices            # source polygon vertices
        self.s_polygonEdges = s_polygonEdges            # source polygon edges
        self.s_matSG        = 'initialShadingGroup'     # source shadingGroup
        
    def __str__( self ):
        return "m_polyIndex = {} s_polyIndex = {} ".format( self.m_polyIndex, self.s_polyIndex )
    
def getPolygonUV( m_fnMesh, m_polygonId, m_vertexIndex, m_uvSet ):
    # pointers to u
    u_util = OpenMaya.MScriptUtil()
    u_util.createFromDouble(0.0)
    u_ptr = u_util.asFloatPtr()
    # pointers to v
    v_util = OpenMaya.MScriptUtil()
    v_util.createFromDouble(0.0)
    v_ptr = v_util.asFloatPtr()
    # pointer to uv index
    id_util = OpenMaya.MScriptUtil()
    id_util.createFromInt(0)
    id_ptr = id_util.asIntPtr()
    #--------------
    m_fnMesh.getPolygonUV( m_polygonId,m_vertexIndex, u_ptr, v_ptr, m_uvSet )
    m_fnMesh.getPolygonUVid( m_polygonId,m_vertexIndex,id_ptr )
    #--------------
    m_u = u_util.getFloatArrayItem(u_ptr,0)
    m_v = v_util.getFloatArrayItem(v_ptr,0)
    m_id = id_util.getIntArrayItem(id_ptr,0)
    return (m_id,m_u,m_v)
    
def getEdgeVertices( m_fnMesh, m_edgeId ):
    m_util = OpenMaya.MScriptUtil() 
    m_util.createFromList([0, 0], 2)
    m_ptr = m_util.asInt2Ptr()
    m_fnMesh.getEdgeVertices( m_edgeId, m_ptr )
    m_start = m_util.getInt2ArrayItem( m_ptr, 0,0 )
    m_end = m_util.getInt2ArrayItem( m_ptr, 0,1 )
    return m_start, m_end
    
def isEdgeSmooth( m_fnMesh, m_start, m_end, m_facesArray ):
    m_state = True
    m_normalStartArr = OpenMaya.MVectorArray()
    m_normalEndArr   = OpenMaya.MVectorArray()
    for m_faceId in m_facesArray:
        m_normalStart = OpenMaya.MVector()
        m_normalEnd   = OpenMaya.MVector()
        m_fnMesh.getFaceVertexNormal( m_faceId, m_start, m_normalStart, OpenMaya.MFn.kWorld )
        m_fnMesh.getFaceVertexNormal( m_faceId, m_end, m_normalEnd, OpenMaya.MFn.kWorld )
        m_normalStartArr.append( m_normalStart )
        m_normalEndArr.append(m_normalEnd)
    m_normalStart1 = m_normalStartArr[0]
    for i in range( m_normalStartArr.length() ):
        m_normalStart2 = m_normalStartArr[i]
        if ( m_normalStart1 != m_normalStart2 ):
            m_state = False
        m_normalStart1 = m_normalStart2
    m_normalEnd1 = m_normalEndArr[0]
    for i in range( m_normalEndArr.length() ):
        m_normalEnd2 = m_normalEndArr[i]
        if ( m_normalEnd1 != m_normalEnd2 ):
            m_state = False
        m_normalEnd1 = m_normalEnd2
    return m_state
    
def getFaceListWithShadingGroupsFromMDagPath( m_path ):
    m_meshFn  = OpenMaya.MFnMesh(m_path)
    m_shaders = OpenMaya.MObjectArray()
    m_faces   = OpenMaya.MIntArray()
    m_meshFn.getConnectedShaders(0, m_shaders, m_faces )
    m_shadersList = []
    m_facesList = []
    for i in range( m_shaders.length() ):
        m_setFn = OpenMaya.MFnSet( m_shaders[i] )
        m_shadersList.append(m_setFn.name())
    for i in range( m_faces.length() ):
        m_facesList.append(m_shadersList[m_faces[i]])
    return m_facesList
    
#--------------------------------------------------------------
def computeTransform(m_point,m_target):
    # input is OpenMaya.MPoint
    # output is point in inverse space of all parents
    m_resMatrix = OpenMaya.MMatrix()
    m_list = []
    getDepth( m_target, m_list )
    for m_item in m_list:
        m_fnDagNode = OpenMaya.MFnDagNode( m_item )
        m_matrix = m_fnDagNode.transformationMatrix() 
        m_resMatrix *= m_matrix
        #m_point = m_point*m_matrix.inverse()
        #print(m_item.fullPathName())
    m_point = m_point*m_resMatrix.inverse()
    return m_point

def getDepth( m_dagPath, m_list ):
    # get the list of parents of the given OpenMaya.MDagPath
    m_list.append(m_dagPath)
    m_dagFn     = OpenMaya.MFnDagNode()
    m_dagFn.setObject( m_dagPath.node() )
    m_objParent     = OpenMaya.MObject()
    m_objParent     = m_dagFn.parent(0)
    if m_objParent.apiType() == OpenMaya.MFn.kWorld:
        return m_list
    else:
        m_dagPathParent = OpenMaya.MDagPath()
        if m_objParent.hasFn( OpenMaya.MFn.kDagNode ):
            OpenMaya.MDagPath.getAPathTo( m_objParent, m_dagPathParent )
            getDepth( m_dagPathParent, m_list )
        return m_list
    